/* a2dos33_img.c version 1.0.0 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "a2dos33_img.h"



#define TRACK_OFFSET(_t)			((_t) * 16 * 256)
#define SECTOR_OFFSET(_t, _s, _o)	((_t) * 16 * 256 + (_o)[(_s)] * 256)



static int get_int2(const unsigned char *data, const unsigned char **next);
static int get_int4(const unsigned char *data, const unsigned char **next);
static void free_a2dos33_ts_list(struct a2dos33_ts_list_t *ts_list);
static void free_a2dos33_catalog(struct a2dos33_catalog_t *catalog);
static int get_a2dos33_ts_list(struct a2dos33_catalog_entry_t *entry, const unsigned char *disk_img, int disk_size, const int *order_table);
static int get_a2dos33_ts_list_all(struct a2dos33_image_t *image);
static int check_a2dos33_catalog_loop(const struct a2dos33_catalog_t *catalog, int track, int sector);
static int get_a2dos33_catalog_entry(struct a2dos33_catalog_entry_t *entry, const unsigned char *src);
static int get_a2dos33_catalog_main(struct a2dos33_catalog_t *catalog, const unsigned char *disk_img, int disk_size, int track, int sector, const int *order_table);
static void get_a2dos33_vtoc(struct a2dos33_vtoc_t *vtoc, const unsigned char *disk_img, int disk_size, const int *order_table);
static int get_a2dos33_catalog(struct a2dos33_image_t *image, int sector_order);
static const struct a2dos33_catalog_entry_t *find_a2dos33_catalog_entry(const struct a2dos33_image_t *image, const char *filename);



static int get_int2(const unsigned char *data, const unsigned char **next)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;

	if (next != NULL) {
		*next = data;
	}

	return val;
}

static int get_int4(const unsigned char *data, const unsigned char **next)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;
	val |= (int)(*data++) << 16;
	val |= (int)(*data++) << 24;

	if (next != NULL) {
		*next = data;
	}

	return val;
}

static void free_a2dos33_ts_list(struct a2dos33_ts_list_t *ts_list)
{
	while (ts_list != NULL) {
		struct a2dos33_ts_list_t *next = ts_list->next;
		free(ts_list);
		ts_list = next;
	}
}

static void free_a2dos33_catalog(struct a2dos33_catalog_t *catalog)
{
	while (catalog != NULL) {
		struct a2dos33_catalog_t *next = catalog->next;
		int i;

		for (i = 0; i < A2DOS33_NUM_CATALOG_ENTRY; i++) {
			free_a2dos33_ts_list(catalog->entry[i].ts_list);
		}

		free(catalog);
		catalog = next;
	}
}

void free_a2dos33_image(struct a2dos33_image_t *image, int free_all)
{
	free_a2dos33_catalog(image->catalog);

	if (image->disk_img != NULL) {
		free(image->disk_img);
	}

	if (free_all) {
		free(image);
	} else {
		memset(image, 0, sizeof(struct a2dos33_image_t));
	}
}

static int get_a2dos33_ts_list(struct a2dos33_catalog_entry_t *entry, const unsigned char *disk_img, int disk_size, const int *order_table)
{
	int result = -1;
	struct a2dos33_track_sector_t ts_list_sector;
	struct a2dos33_ts_list_t *ts_list;
	struct a2dos33_ts_list_t *last = NULL;
	int remain = entry->num_sectors;

	ts_list_sector = entry->first_ts_list_sector;
	if (ts_list_sector.track == 0xFF) {
		return 0;
	}

	do {
		const unsigned char *src;
		int offset;
		int i;

		if ((ts_list = (struct a2dos33_ts_list_t *)malloc(sizeof(struct a2dos33_ts_list_t))) == NULL) {
			printf("memory exhaust!\n");
			goto ERROR_EXIT;
		}
		memset(ts_list, 0, sizeof(struct a2dos33_ts_list_t));

		if (last == NULL) {
			entry->ts_list = ts_list;
		} else {
			last->next = ts_list;
		}
		last = ts_list;

		offset = SECTOR_OFFSET(ts_list_sector.track, ts_list_sector.sector, order_table);
		if (disk_size < offset + 256) {
			printf("illegal track/sector (%02X/%02X)\n",
				   ts_list_sector.track,
				   ts_list_sector.sector);
			goto ERROR_EXIT;
		}
		src = disk_img + offset;

		ts_list->unused1 = *src++;
		ts_list->next_ts_list_sector.track = *src++;
		ts_list->next_ts_list_sector.sector = *src++;
		ts_list->unused2[0] = *src++;
		ts_list->unused2[1] = *src++;
		ts_list->sector_number_in_file = get_int2(src, &src);
		memcpy(ts_list->unused3, src, 5);
		src += 5;

		for (i = 0; i < A2DOS33_NUM_TS_LIST; i++) {
			ts_list->list[i].track = *src++;
			ts_list->list[i].sector = *src++;
			if (0 < remain) {
				remain--;
				if (disk_size < SECTOR_OFFSET(ts_list->list[i].track, ts_list->list[i].sector, order_table) + 256) {
					printf("illegal track/sector list[%02X:%02X] (%02X/%02X)\n",
						   ts_list->sector_number_in_file,
						   i,
						   ts_list->list[i].track,
						   ts_list->list[i].sector);
				}
			}
		}

		ts_list_sector = ts_list->next_ts_list_sector;
	} while (ts_list_sector.track != 0);

	result = 0;

ERROR_EXIT:
	if (result < 0) {
		free_a2dos33_ts_list(entry->ts_list);
		entry->ts_list = NULL;
	}

	return result;
}

static int get_a2dos33_ts_list_all(struct a2dos33_image_t *image)
{
	int result = -1;
	struct a2dos33_catalog_t *catalog;
	const int *order_table = image->sector_order == A2DSK_SECTOR_ORDER_PRODOS ? sector_order_table_prodos : sector_order_table_dos33;

	for (catalog = image->catalog; catalog != NULL; catalog = catalog->next) {
		int i;
		for (i = 0; i < A2DOS33_NUM_CATALOG_ENTRY; i++) {
			if (catalog->entry[i].num_sectors == 0) {
				continue;
			}
			if (get_a2dos33_ts_list(catalog->entry + i,
								  image->disk_img,
								  image->disk_size,
								  order_table) < 0) {
				goto ERROR_EXIT;
			}
		}
	}

	result = 0;

ERROR_EXIT:

	return result;
}

static int check_a2dos33_catalog_loop(const struct a2dos33_catalog_t *catalog, int track, int sector)
{
	for ( ; catalog != NULL; catalog = catalog->next) {
		if (catalog->next_catalog_sector.track == track &&
			catalog->next_catalog_sector.sector == sector &&
			catalog->next != NULL) {
/*
			printf("found catalog loop\n");
*/
			return -1;
		}
	}
	return 0;
}

static int get_a2dos33_catalog_entry(struct a2dos33_catalog_entry_t *entry, const unsigned char *src)
{
	entry->first_ts_list_sector.track = *src++;
	entry->first_ts_list_sector.sector = *src++;
	entry->file_type = *src++;
	memcpy(entry->filename, src, A2DOS33_FILENAME_LEN);
	src += A2DOS33_FILENAME_LEN;
	entry->num_sectors = get_int2(src, &src);

	return 0;
}

static int get_a2dos33_catalog_main(struct a2dos33_catalog_t *catalog, const unsigned char *disk_img, int disk_size, int track, int sector, const int *order_table)
{
	const unsigned char *src;
	int offset;
	int i;

	if (track <= 0 || A2DOS33_NUM_TRACKS <= track || sector < 0 || 16 <= sector) {
		return -1;
	}

	offset = SECTOR_OFFSET(track, sector, order_table);
	if (disk_size < offset + 256) {
		return -1;
	}
	src = disk_img + offset;

	catalog->unused1 = *src++;
	catalog->next_catalog_sector.track = *src++;
	catalog->next_catalog_sector.sector = *src++;
	memcpy(catalog->unused2, src, 8);
	src += 8;

	for (i = 0; i < A2DOS33_NUM_CATALOG_ENTRY; i++) {
		if (get_a2dos33_catalog_entry(catalog->entry + i, src) < 0) {
			return -1;
		}
		src += 35;
	}

	return 0;
}

static void get_a2dos33_vtoc(struct a2dos33_vtoc_t *vtoc, const unsigned char *disk_img, int disk_size, const int *order_table)
{
	const unsigned char *src;
	int i;

	memset(vtoc, 0, sizeof(struct a2dos33_vtoc_t));

	src = disk_img + SECTOR_OFFSET(A2DOS33_VTOC_TRACK, A2DOS33_VTOC_SECTOR, order_table);

	vtoc->unused1 = *src++;
	vtoc->catalog_sector.track = *src++;
	vtoc->catalog_sector.sector = *src++;
	vtoc->dos_version = *src++;
	vtoc->unused2[0] = *src++;
	vtoc->unused2[1] = *src++;
	vtoc->disk_volume = *src++;
	memcpy(vtoc->unused3, src, 32);
	src += 32;
	vtoc->num_ts_pairs = *src++;
	memcpy(vtoc->unused4, src, 8);
	src += 8;
	vtoc->last_track = *src++;
	vtoc->dir_alloc = *src++;
	vtoc->unused5[0] = *src++;
	vtoc->unused5[1] = *src++;
	vtoc->num_tracks_per_disk = *src++;
	vtoc->num_sectors_per_track = *src++;
	vtoc->num_bytes_per_sector = (unsigned short)get_int2(src, &src);

	for (i = 0; i < A2DOS33_NUM_TRACKS; i++) {
		vtoc->track_bitmap[i] = (unsigned int)get_int4(src, &src);
	}
}

static int get_a2dos33_catalog(struct a2dos33_image_t *image, int sector_order)
{
	int result = -1;
	struct a2dos33_catalog_t *catalog;
	struct a2dos33_catalog_t *last = NULL;
	struct a2dos33_track_sector_t catalog_sector = {
		A2DOS33_CATALOG_START_TRACK,
		A2DOS33_CATALOG_START_SECTOR
	};

	if (image->num_tracks <= A2DOS33_VTOC_TRACK) {
		printf("disk image size error\n");
		return -1;
	}

	if (sector_order == A2DSK_SECTOR_ORDER_DOS33) {
		image->sector_order_table = sector_order_table_dos33;
	} else if (sector_order == A2DSK_SECTOR_ORDER_PRODOS) {
		image->sector_order_table = sector_order_table_prodos;
	} else {
/*
		printf("sector order not specified\n");
		printf("try DOS 3.3 order\n");
*/
		if (get_a2dos33_catalog(image, A2DSK_SECTOR_ORDER_DOS33) == 0) {
			image->sector_order = A2DSK_SECTOR_ORDER_DOS33;
			image->sector_order_table = sector_order_table_dos33;
			return 0;
		}
/*
		printf("try ProDOS order\n");
*/
		if (get_a2dos33_catalog(image, A2DSK_SECTOR_ORDER_PRODOS) == 0) {
			image->sector_order = A2DSK_SECTOR_ORDER_PRODOS;
			image->sector_order_table = sector_order_table_prodos;
			return 0;
		}
		image->sector_order_table = sector_order_table_dos33;
		return -1;
	}

	get_a2dos33_vtoc(&(image->vtoc), image->disk_img, image->disk_size, image->sector_order_table);

	do {
		if (check_a2dos33_catalog_loop(image->catalog, catalog_sector.track, catalog_sector.sector) < 0) {
			goto ERROR_EXIT;
		}

		if ((catalog = (struct a2dos33_catalog_t *)malloc(sizeof(struct a2dos33_catalog_t))) == NULL) {
			printf("memory exhaust!\n");
			goto ERROR_EXIT;
		}
		memset(catalog, 0, sizeof(struct a2dos33_catalog_t));

		if (last == NULL) {
			image->catalog = catalog;
		} else {
			last->next = catalog;
		}
		last = catalog;

		if (get_a2dos33_catalog_main(catalog,
								   image->disk_img,
								   image->disk_size,
								   catalog_sector.track,
								   catalog_sector.sector,
								   image->sector_order_table) < 0) {
			goto ERROR_EXIT;
		}
		catalog_sector = catalog->next_catalog_sector;
	} while (catalog_sector.track != 0);

	result = 0;

ERROR_EXIT:
	if (result < 0) {
		free_a2dos33_catalog(image->catalog);
		image->catalog = NULL;
	}

	return result;
}

int read_a2dos33_image(struct a2dos33_image_t *image, const char *image_name, int sector_order)
{
	int result = -1;
	FILE *fp = NULL;

	memset(image, 0, sizeof(struct a2dos33_image_t));
	image->sector_order = sector_order;
	image->sector_order_table = sector_order_table_dos33;

	if ((fp = fopen(image_name, "rb")) == NULL) {
		printf("open error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("seek error '%s' (SEEK_END)\n", image_name);
		goto ERROR_EXIT;
	}
	image->disk_size = ftell(fp);
	image->num_tracks = image->disk_size / (16 * 256);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("seek error '%s' (SEEK_SET)\n", image_name);
		goto ERROR_EXIT;
	}

	if ((image->disk_img = (unsigned char *)malloc(image->disk_size)) == NULL) {
		printf("memory exhaust!\n");
		goto ERROR_EXIT;
	}

	if (fread(image->disk_img, image->disk_size, 1, fp) != 1) {
		printf("read error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if (get_a2dos33_catalog(image, sector_order) < 0) {
		printf("can't get catalog '%s'\n", image_name);
		goto ERROR_EXIT;
	}
	if (get_a2dos33_ts_list_all(image) < 0) {
		printf("can't get track/sector list '%s'\n", image_name);
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

int read_a2dos33_image_file(const struct a2dos33_image_t *image, const char *filename, unsigned char *data, int data_size)
{
	int read_size = 0;
	const struct a2dos33_catalog_entry_t *entry = find_a2dos33_catalog_entry(image, filename);
	const struct a2dos33_ts_list_t *ts_list;
	int ts_list_index;
	int num_sectors;
	const unsigned char *sector_data;
	int file_type;
	int file_size;
	int sector_data_size;

	if (entry == NULL) {
		printf("file not found '%s'\n", filename);
		return -1;
	}

	memset(data, 0, data_size);

	ts_list = entry->ts_list;
	num_sectors = entry->num_sectors;
	num_sectors--;

	if (ts_list == NULL || num_sectors <= 0) {
		return -1;
	}

	sector_data = image->disk_img + SECTOR_OFFSET(ts_list->list[0].track,
												  ts_list->list[0].sector,
												  image->sector_order_table);
	num_sectors--;
	sector_data_size = 256;

	file_type = entry->file_type & 0x7F;
	if (file_type == 0x01 || file_type == 0x02) {
		/* Integer Basic, Applesoft Basic */
		file_size = get_int2(sector_data, NULL);
		sector_data += 2;
		sector_data_size -= 2;
	} else if (file_type == 0x04) {
		/* Binary */
		file_size = get_int2(sector_data + 2, NULL);
		sector_data += 4;
		sector_data_size -= 4;
	} else if (file_type == 0x10) {
		/* Relocatable */
		file_size = get_int2(sector_data + 2, NULL);
		sector_data += 6;
		sector_data_size -= 6;
	} else {
		/* Text, S, A, B */
		file_size = -1;
	}

	if (0 <= file_size && file_size != data_size) {
		printf("file size mismatch '%s' %d != %d\n", filename, file_size, data_size);
	}

	while (0 < data_size) {
		for (ts_list_index = 0; ts_list_index < A2DOS33_NUM_TS_LIST; ts_list_index++) {
			int len;

			if (0 < read_size) {
				if (num_sectors <= 0) {
					return read_size;
				}

				sector_data = image->disk_img + SECTOR_OFFSET(ts_list->list[ts_list_index].track,
															  ts_list->list[ts_list_index].sector,
															  image->sector_order_table);
				num_sectors--;
				sector_data_size = 256;
			}

			if (0 <= file_size && file_size < sector_data_size) {
				sector_data_size = file_size;
			}

			if (data_size < sector_data_size) {
				len = data_size;
			} else {
				len = sector_data_size;
			}

			memcpy(data, sector_data, len);
			data += len;
			read_size += len;
			data_size -= len;
			if (data_size <= 0) {
				return read_size;
			}
			if (0 < file_size) {
				file_size -= len;
				if (file_size <= 0) {
					return read_size;
				}
			}
		}

		ts_list = ts_list->next;
		num_sectors--;
		if (num_sectors <= 0) {
			break;
		}
	}

	return read_size;
}

static const struct a2dos33_catalog_entry_t *find_a2dos33_catalog_entry(const struct a2dos33_image_t *image, const char *filename)
{
	const struct a2dos33_catalog_t *catalog;
	char entry_name[A2DOS33_FILENAME_LEN];
	int i;

	for (i = 0; i < A2DOS33_FILENAME_LEN && filename[i] != '\0'; i++) {
		entry_name[i] = (filename[i] | 0x80);
	}
	for ( ; i < A2DOS33_FILENAME_LEN; i++) {
		entry_name[i] = 0xA0;
	}

	for (catalog = image->catalog; catalog != NULL; catalog = catalog->next) {
		for (i = 0; i < A2DOS33_NUM_CATALOG_ENTRY; i++) {
			if (memcmp(catalog->entry[i].filename, entry_name, A2DOS33_FILENAME_LEN) == 0) {
				return (catalog->entry + i);
			}
		}
	}

	return NULL;
}

/* EOF */
