/* dosfd_img.c version 1.0.1 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dosfd_img.h"


static int get_int2(const unsigned char *data);
static int get_int4(const unsigned char *data);
static int *get_fat(const unsigned char *buf, int size, int offset, int *fat_size);
static int make_entry_name(char *entry_name, const char *file_name);


static int get_int2(const unsigned char *data)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;

	return val;
}

static int get_int4(const unsigned char *data)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;
	val |= (int)(*data++) << 16;
	val |= (int)(*data++) << 24;

	return val;
}

static int *get_fat(const unsigned char *buf, int size, int offset, int *fat_size)
{
	int *table;
	int i;

	*fat_size = (size - offset) / 0x400;
	table = (int *)malloc(sizeof(int) * (*fat_size));
	if (table == NULL) {
		printf("memory exhaust!");
		return NULL;
	}

	offset += 0x400;
	for (i = 0; i < (*fat_size); i += 2) {
		int val;
		val = (int)buf[offset++];
		val |= ((int)buf[offset++]) << 8;
		val |= ((int)buf[offset++]) << 16;
		table[i    ] = val & 0xFFF;
		table[i + 1] = val >> 12;
	}

	return table;
}

static int make_entry_name(char *entry_name, const char *file_name)
{
	const char *src = file_name;
	int i;

	for (i = 0; i < 8 && *src && *src != '.'; i++, src++) {
		entry_name[i] = *src;
	}
	for ( ; i < 8; i++) {
		entry_name[i] = ' ';
	}

	if (*src == '.') {
		src++;
		for ( ; i < 11 && *src; i++, src++) {
			entry_name[i] = *src;
		}
	}
	for ( ; i < 11; i++) {
		entry_name[i] = ' ';
	}
	entry_name[i] = '\0';

	if (*src) {
		return -1;
	}

	return 0;
}

void free_dosfd_image(struct dosfd_image_t *image, int free_all)
{
	if (image->disk_img != NULL) {
		free(image->disk_img);
	}

	if (image->fat != NULL) {
		free(image->fat);
	}

	if (free_all) {
		free(image);
	} else {
		memset(image, 0, sizeof(struct dosfd_image_t));
	}
}

int read_dosfd_image(struct dosfd_image_t *image, const char *image_name, int img_offset)
{
	int result = -1;
	FILE *fp = NULL;

	memset(image, 0, sizeof(*image));
	image->img_offset = img_offset;

	if ((fp = fopen(image_name, "rb")) == NULL) {
		printf("open error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("seek error '%s' (SEEK_END)\n", image_name);
		goto ERROR_EXIT;
	}
	image->disk_size = ftell(fp);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("seek error '%s' (SEEK_SET)\n", image_name);
		goto ERROR_EXIT;
	}

	if ((image->disk_img = (unsigned char *)malloc(image->disk_size)) == NULL) {
		printf("memory exhaust!\n");
		goto ERROR_EXIT;
	}

	if (fread(image->disk_img, image->disk_size, 1, fp) != 1) {
		printf("read error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if ((image->fat = get_fat(image->disk_img, image->disk_size, img_offset, &(image->fat_size))) == NULL) {
		printf("can't get fat '%s' offset=%08X\n", image_name, img_offset);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (result < 0) {
		free_dosfd_image(image, 0);
	}

	return result;
}

int read_dosfd_image_file(const struct dosfd_image_t *image, const char *file_name, unsigned char *data, int data_size, int ignore_case)
{
	char entry_name[12];
	int entry_size;
	int cluster;
	int offset;
	int found = 0;

	if (make_entry_name(entry_name, file_name) < 0) {
		printf("illegal filename '%s'\n", file_name);
		return -1;
	}

	for (offset = 0x1400 + image->img_offset; offset < 0x2C00 + image->img_offset; offset += 0x20) {
		if ((ignore_case == 0 && strncmp((const char *)(image->disk_img + offset), entry_name, 11) == 0) ||
			(ignore_case != 0 && strncasecmp((const char *)(image->disk_img + offset), entry_name, 11) == 0)) {
			cluster = get_int2(image->disk_img + offset + 0x1A);
			entry_size = get_int4(image->disk_img + offset + 0x1C);
			found = 1;
			break;
		}
	}
	if (! found) {
		printf("can't find '%s'\n", file_name);
		return -1;
	}

	if (entry_size < data_size) {
		printf("file size error '%s' %d < %d\n", file_name, entry_size, data_size);
		return -1;
	} else if (data_size < entry_size) {
		printf("file size mismatch '%s' %d != %d ... ignored\n", file_name, entry_size, data_size);
	}

	{
		unsigned char *des = data;
		int size;

		for (size = data_size; 0 < size; size -= 0x400, des += 0x400, cluster = image->fat[cluster]) {
			if (cluster < 2 || image->fat_size <= cluster) {
				printf("fat error '%s'\n", file_name);
				return -1;
			}
			memcpy(des, image->disk_img + cluster * 0x400 + 0x2400 + image->img_offset, (0x400 < size ? 0x400 : size));
		}
	}

	return 0;
}

int write_dosfd_image(const struct dosfd_image_t *image, const char *image_name)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(image_name, "wb")) == NULL) {
		printf("open error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	if (fwrite(image->disk_img, image->disk_size, 1, fp) != 1) {
		printf("write error '%s'\n", image_name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

int write_dosfd_image_file(struct dosfd_image_t *image, const char *file_name, const unsigned char *data, int data_size, int ignore_case)
{
	char entry_name[12];
	int entry_size;
	int cluster;
	int offset;
	int found = 0;

	if (make_entry_name(entry_name, file_name) < 0) {
		printf("illegal filename '%s'\n", file_name);
		return -1;
	}

	for (offset = 0x1400 + image->img_offset; offset < 0x2C00 + image->img_offset; offset += 0x20) {
		if ((ignore_case == 0 && strncmp((const char *)(image->disk_img + offset), entry_name, 11) == 0) ||
			(ignore_case != 0 && strncasecmp((const char *)(image->disk_img + offset), entry_name, 11) == 0)) {
			cluster = get_int2(image->disk_img + offset + 0x1A);
			entry_size = get_int4(image->disk_img + offset + 0x1C);
			found = 1;
			break;
		}
	}
	if (! found) {
		printf("can't find '%s'\n", file_name);
		return -1;
	}

	if (entry_size < data_size) {
		printf("file size error '%s' %d < %d\n", file_name, entry_size, data_size);
		return -1;
	} else if (data_size < entry_size) {
		printf("file size mismatch '%s' %d != %d ... ignored\n", file_name, entry_size, data_size);
	}

	{
		const unsigned char *src = data;
		int size;

		for (size = data_size; 0 < size; size -= 0x400, src += 0x400, cluster = image->fat[cluster]) {
			if (cluster < 2 || image->fat_size <= cluster) {
				printf("fat error '%s'\n", file_name);
				return -1;
			}
			memcpy(image->disk_img + cluster * 0x400 + 0x2400 + image->img_offset, src, (0x400 < size ? 0x400 : size));
		}
	}

	return 0;
}

/* EOF */
