#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "bitmap.h"
#include "dosfd_img.h"
#include "u4pc98shape_info.h"


static const struct bitmap_palette_t shape_palette[16] = {
	RGB(0x00, 0x00, 0x00),		/* black */
	RGB(0x00, 0x00, 0xFF),		/* blue */
	RGB(0xFF, 0x00, 0x00),		/* red */
	RGB(0xFF, 0x00, 0xFF),		/* magenta */
	RGB(0x00, 0xFF, 0x00),		/* green */
	RGB(0x00, 0xFF, 0xFF),		/* cyan */
	RGB(0xFF, 0xFF, 0x00),		/* yellow */
	RGB(0xFF, 0xFF, 0xFF),		/* white */
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00)
};

static const struct bitmap_palette_t mask_palette[2] = {
	RGB(0x00, 0x00, 0x00),		/* black */
	RGB(0xFF, 0xFF, 0xFF)		/* white */
};

static struct dosfd_image_t ultima_disk_image;
static struct dosfd_image_t britannia_disk_image;
static unsigned char ocshape_data[108160];
static unsigned char swshape_data[62464];
static unsigned char vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];
static unsigned char mask_vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];


static void usage(void);
static char *create_write_name(const char *name);
static int read_shapes_bmp(const char *bmp_file, int *rows, int bit_count, const struct bitmap_palette_t *palette, unsigned char *vram);
static void merge_shapes_and_mask(unsigned char *vram, const unsigned char *mask_vram, int size);
static int change_shape(int row, int col, unsigned char *data, int bit_count, const unsigned char *vram);
static int confirm(const char *msg);



static void usage(void)
{
	printf("usage: u4pc98chg_shape [options] SHAPES.BMP [MASK.BMP]\n");
	printf("option: --disk1 DISK-IMG    disk image name of Ultima Disk\n");
	printf("                            read 'U4_Ultima.fdi' if not specified\n");
	printf("        --disk2 DISK-IMG    disk image name of Britannia Disk\n");
	printf("                            read 'U4_Britania.fdi' if not specified\n");
	printf("        --write1 DISK-IMG   disk image name of modified Ultima Disk\n");
	printf("        --write2 DISK-IMG   disk image name of modified Britannia Disk\n");
	printf("        --no-shores         remove shore line shapes\n");
	printf("        --help              print usage\n");
	printf("\n");
	printf("remove all mask if mask bitmap is not specified\n");
	printf("supported disk image format is *.FDI only\n");
	printf("write modified disk image to filename.NEW.ext if not specified\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *bmp_file = NULL;
	const char *mask_file = NULL;

	struct param_info {
		const char *disk_name;
		const char *write_name;
		const char *file_name;
		unsigned char *shape_data;
		int data_size;
		struct dosfd_image_t *image;
	} params[] = {
		{ "U4_Ultima.fdi"  , NULL, "ocshape.pat", ocshape_data, sizeof(ocshape_data), &ultima_disk_image },
		{ "U4_Britania.fdi", NULL, "swshape.pat", swshape_data, sizeof(swshape_data), &britannia_disk_image }
	};
	unsigned char *shape_data[3] = {
		NULL,
		ocshape_data,
		swshape_data
	};
	int no_shores = 0;
	int bmp_rows;
	int mask_rows;
	int row;
	int col;
	int i;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk1") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (ultima disk)\n");
					return 1;
				}
				params[0].disk_name = *argv;
			} else if (strcmp(*argv, "--disk2") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (britannia disk)\n");
					return 1;
				}
				params[1].disk_name = *argv;
			} else if (strcmp(*argv, "--write1") == 0) {
				if (*++argv == NULL) {
					printf("no disk image to write (ultima disk)\n");
					return 1;
				}
				params[0].write_name = *argv;
			} else if (strcmp(*argv, "--write2") == 0) {
				if (*++argv == NULL) {
					printf("no disk image to write (britannia disk)\n");
					return 1;
				}
				params[1].write_name = *argv;
			} else if (strcmp(*argv, "--no-shores") == 0) {
				no_shores = 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				bmp_file = *argv;
			} else {
				mask_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (bmp_file == NULL) {
		printf("no shapes bitmap filename\n");
		return 1;
	}

	for (i = 0; i < 2; i++) {
		if (params[i].write_name == NULL) {
			params[i].write_name = create_write_name(params[i].disk_name);
			if (params[i].write_name == NULL) {
				printf("can't create filename\n");
				return 1;
			}
		}
	}

	memset(vram, 0, sizeof(vram));
	memset(mask_vram, 0, sizeof(mask_vram));

	/* read shapes bitmap */
	if (read_shapes_bmp(bmp_file, &bmp_rows, 4, shape_palette, vram) < 0) {
		printf("read shapes file error '%s'\n", bmp_file);
		return 1;
	}

	/* read mask bitmap */
	if (mask_file != NULL) {
		if (read_shapes_bmp(mask_file, &mask_rows, 1, mask_palette, mask_vram) < 0) {
			printf("read mask file error '%s'\n", mask_file);
			return 1;
		}
		if (bmp_rows != mask_rows) {
			printf("shapes and mask size mismatch\n");
			return 1;
		}
	}

	/* remove shore line shapes */
	if (no_shores) {
		memset(vram      + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 17, 0x00, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS);
		memset(vram      + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 19, 0x00, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 2);
		memset(mask_vram + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 17, 0xFF, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS);
		memset(mask_vram + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 19, 0xFF, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 2);
	}

	/* merge shapes and mask */
	merge_shapes_and_mask(vram, mask_vram, sizeof(vram));

	/* read disk images */
	for (i = 0; i < 2; i++) {
		if (read_dosfd_image(params[i].image, params[i].disk_name, FDI_OFFSET) < 0) {
			printf("read disk image error %s\n", params[i].disk_name);
			return 1;
		}
		if (read_dosfd_image_file(params[i].image, params[i].file_name, params[i].shape_data, params[i].data_size, 1) < 0) {
			printf("read disk image error %s/%s\n", params[i].disk_name, params[i].file_name);
			return 1;
		}
	}

	/* change shapes */
	for (row = 0; row < NUM_ROWS; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			int src_row = row;
			int src_col = col;

			if (shape_info[row][col].file_id == FILE_ID_NONE) {
				continue;
			}

			if (bmp_rows <= row) {
				src_row = shape_info[row][col].shape / NUM_COLS;
				src_col = shape_info[row][col].shape % NUM_COLS;

				if (bmp_rows <= src_row && ! no_shores) {
					continue;
				}
			}
			change_shape(src_row, src_col, shape_data[shape_info[row][col].file_id] + shape_info[row][col].offset, shape_info[row][col].bit_count, vram);
		}
	}

	/* write modified disk images */
	for (i = 0; i < 2; i++) {
		struct stat st;

		if (stat(params[i].write_name, &st) == 0) {
			printf("already exists '%s'\n", params[i].write_name);
			if (confirm("overwrite ?") <= 0) {
				return 1;
			}
		}

		if (write_dosfd_image_file(params[i].image, params[i].file_name, params[i].shape_data, params[i].data_size, 1) < 0) {
			printf("write disk image error %s/%s\n", params[i].write_name, params[i].file_name);
			return 1;
		}
		if (write_dosfd_image(params[i].image, params[i].write_name) < 0) {
			printf("write disk image error %s\n", params[i].write_name);
			return 1;
		}
		free_dosfd_image(params[i].image, 0);

		printf("write '%s'\n", params[i].write_name);
	}

	return 0;
}

static char *create_write_name(const char *name)
{
	int baselen;
	const char *ext;
	int len = strlen(name) + 5;
	char *new_name = (char *)malloc(len);
	if (new_name == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}

	if ((ext = strrchr(name, '.')) == NULL) {
		baselen = strlen(name);
		ext = "";
	} else {
		baselen = ext - name;
	}

	strncpy(new_name, name, baselen);
	strcpy(new_name + baselen, ".NEW");
	strcpy(new_name + baselen + 4, ext);

	return new_name;
}

static int read_shapes_bmp(const char *name, int *rows, int bit_count, const struct bitmap_palette_t *palette, unsigned char *vram)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct bitmap_palette_t *work_palette = NULL;
	unsigned char *work_data = NULL;

	if ((fp = fopen(name, "rb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (read_bitmap(fp, &header, &work_palette, &work_data, 0) < 0) {
		printf("read error '%s'\n", name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if (header.width != SHAPE_WIDTH * NUM_COLS ||
		header.height < SHAPE_HEIGHT * NUM_MIN_ROWS || SHAPE_HEIGHT * NUM_ROWS < header.height) {
		printf("bitmap size error '%s' %d x %d\n", name, header.width, header.height);
		goto ERROR_EXIT;
	}

	*rows = header.height / SHAPE_HEIGHT;

	if (convert_bit_count(header.width, header.height,
						  header.bit_count, work_palette, work_data,
						  bit_count, palette, &vram) < 0) {
		printf("convert_bit_count() error '%s'\n", name);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (work_palette != NULL) {
		free(work_palette);
	}
	if (work_data != NULL) {
		free(work_data);
	}

	return result;
}

static void merge_shapes_and_mask(unsigned char *vram, const unsigned char *mask_vram, int size)
{
	while (0 < size--) {
		if (*mask_vram) {
			*vram |= 0x08;
		}
		mask_vram++;
		vram++;
	}
}

static int change_shape(int row, int col, unsigned char *data, int bit_count, const unsigned char *vram)
{
	int plane;
	int y;
	int x;
	int pal;

	for (plane = 0; plane < bit_count; plane++) {
		if (bit_count == 3) {
			pal = (1 << plane);
		} else {
			pal = (1 << ((plane + 3) & 3));
		}

		for (y = 0; y < SHAPE_HEIGHT; y++) {
			const unsigned char *src = vram + (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;

			for (x = 0; x < SHAPE_WIDTH; x += 8) {
				unsigned char val = 0;
				int bit;

				for (bit = 0; bit < 8; bit++, src++) {
					const int mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

					if (*src & pal) {
						val |= mask[bit];
					}
				}

				*data++ = val;
			}
		}
	}

	return 0;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
