#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "bitmap.h"
#include "dosfd_img.h"
#include "u4x68kshape_info.h"


static const struct bitmap_palette_t palette[16] = {
	RGB(0x00, 0x00, 0x00),		/* black */
	RGB(0x00, 0x00, 0xFF),		/* blue */
	RGB(0xFF, 0x00, 0x00),		/* red */
	RGB(0xFF, 0x00, 0xFF),		/* magenta */
	RGB(0x00, 0xFF, 0x00),		/* green */
	RGB(0x00, 0xFF, 0xFF),		/* cyan */
	RGB(0xFF, 0xFF, 0x00),		/* yellow */
	RGB(0xFF, 0xFF, 0xFF),		/* white */
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00)
};

static const struct bitmap_palette_t mask_palette[16] = {
	RGB(0x00, 0x00, 0x00),		/* black */
	RGB(0xFF, 0xFF, 0xFF),		/* white */
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00)
};

static unsigned char shape_data[212544];
static unsigned char vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];
static unsigned char mask_vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];


static void usage(void);
static int read_shape_bmp(const char *bmp_file, int *rows, int *cols, const struct bitmap_palette_t *palette, unsigned char *vram);
static int change_shape(int row, int col, unsigned char *data, const unsigned char *vram, const unsigned char *mask_vram);
static int confirm(const char *msg);



static void usage(void)
{
	printf("usage: u4x68kchg_shape [options] SHAPES.BMP [MASK.BMP]\n");
	printf("option: --disk DISK-IMG     disk image name of Ultima Disk\n");
	printf("                            read 'u4-ultima.xdf' if not specified\n");
	printf("        --write DISK-IMG    disk image name of modified Ultima Disk\n");
	printf("        --no-shores         remove shore line shapes\n");
	printf("        --help              print usage\n");
	printf("\n");
	printf("remove all mask if mask bitmap is not specified\n");
	printf("supported disk image format is *.XDF only\n");
	printf("write modified disk image to filename.NEW.ext if not specified\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *bmp_file = NULL;
	const char *mask_file = NULL;

	const char *disk_name = "u4-ultima.xdf";
	const char *write_name = "u4-ultima.NEW.xdf";
	const char *file_name = "shape.pat";
	int no_shores = 0;
	int bmp_rows;
	int bmp_cols;
	int mask_rows;
	int mask_cols;
	int row;
	int col;
	struct dosfd_image_t image;
	struct stat st;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (ultima disk)\n");
					return 1;
				}
				disk_name = *argv;
			} else if (strcmp(*argv, "--write") == 0) {
				if (*++argv == NULL) {
					printf("no disk image to write (ultima disk)\n");
					return 1;
				}
				write_name = *argv;
			} else if (strcmp(*argv, "--file") == 0) {
				if (*++argv == NULL) {
					printf("no shape file name\n");
					return 1;
				}
				file_name = *argv;
			} else if (strcmp(*argv, "--no-shores") == 0) {
				no_shores = 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				bmp_file = *argv;
			} else {
				mask_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (bmp_file == NULL) {
		printf("no shapes bitmap filename\n");
		return 1;
	}

	memset(vram, 0, sizeof(vram));
	memset(mask_vram, 0, sizeof(mask_vram));

	if (read_shape_bmp(bmp_file, &bmp_rows, &bmp_cols, palette, vram) < 0) {
		printf("read bitmap file error '%s'\n", bmp_file);
		return 1;
	}

	if (mask_file != NULL) {
		if (read_shape_bmp(mask_file, &mask_rows, &mask_cols, mask_palette, mask_vram) < 0) {
			printf("read mask file error '%s'\n", mask_file);
			return 1;
		}
		if (bmp_rows != mask_rows || bmp_cols != mask_cols) {
			printf("bitmap and mask size mismatch\n");
			return 1;
		}
	}

	if (no_shores) {
		memset(vram      + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 17, 0x00, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS);
		memset(vram      + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 19, 0x00, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 2);
		memset(mask_vram + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 17, 0xFF, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS);
		memset(mask_vram + SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 19, 0xFF, SHAPE_WIDTH * SHAPE_HEIGHT * NUM_COLS * 2);
	}

	if (read_dosfd_image(&image, disk_name, 0) < 0) {
		printf("read disk image error %s\n", disk_name);
		return 1;
	}
	if (read_dosfd_image_file(&image, file_name, shape_data, sizeof(shape_data), 1) < 0) {
		printf("read disk image error %s/%s\n", disk_name, file_name);
		return 1;
	}

	for (row = 0; row < NUM_ROWS; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			int src_row = row;
			int src_col = col;

			if (shape_info[row][col].shape < 0) {
				continue;
			}

			if (bmp_rows <= row) {
				src_row = shape_info[row][col].shape / NUM_COLS;
				src_col = shape_info[row][col].shape % NUM_COLS;

				if (bmp_rows <= src_row && ! no_shores) {
					continue;
				}
			}
			change_shape(src_row, src_col, shape_data + shape_info[row][col].offset, vram, mask_vram);
		}
	}

	if (stat(write_name, &st) == 0) {
		printf("already exists '%s'\n", write_name);
		if (confirm("overwrite ?") <= 0) {
			return 1;
		}
	}

	if (write_dosfd_image_file(&image, file_name, shape_data, sizeof(shape_data), 1) < 0) {
		printf("write disk image error %s/%s\n", disk_name, file_name);
		return 1;
	}
	if (write_dosfd_image(&image, write_name) < 0) {
		printf("write disk image error %s\n", disk_name);
		return 1;
	}
	free_dosfd_image(&image, 0);

	return 0;
}

static int read_shape_bmp(const char *name, int *rows, int *cols, const struct bitmap_palette_t *palette, unsigned char *vram)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct bitmap_palette_t *work_palette = NULL;
	unsigned char *work_data = NULL;

	if ((fp = fopen(name, "rb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (read_bitmap(fp, &header, &work_palette, &work_data, 0) < 0) {
		printf("read error '%s'\n", name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if (header.width != SHAPE_WIDTH * NUM_COLS || header.height < SHAPE_HEIGHT * NUM_MIN_ROWS || SHAPE_HEIGHT * NUM_ROWS < header.height) {
		printf("bitmap size error '%s' %d x %d\n", name, header.width, header.height);
		goto ERROR_EXIT;
	}

	*rows = header.height / SHAPE_HEIGHT;
	*cols = header.width / SHAPE_WIDTH;

	if (convert_bit_count(header.width, header.height,
						  header.bit_count, work_palette, work_data,
						  4, palette, &vram) < 0) {
		printf("convert_bit_count() error '%s'\n", name);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (work_palette != NULL) {
		free(work_palette);
	}
	if (work_data != NULL) {
		free(work_data);
	}

	return result;
}

static int change_shape(int row, int col, unsigned char *data, const unsigned char *vram, const unsigned char *mask_vram)
{
	int y;
	int x;

	for (y = 0; y < SHAPE_HEIGHT; y++) {
		const unsigned char *src = vram + (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;
		const unsigned char *mask_src = mask_vram + (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;

		for (x = 0; x < SHAPE_WIDTH; x += 2) {
			unsigned char val = 0;

			val = (*src++) << 4;
			if (*mask_src++) {
				val |= 0x80;
			}
			val |= (*src++) & 0x0F;
			if (*mask_src++) {
				val |= 0x08;
			}

			*data++ = val;
		}
	}

	return 0;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
