#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "bitmap.h"
#include "dosfd_img.h"
#include "u4x68kshape_info.h"


static const struct bitmap_palette_t palette[16] = {
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0xFF, 0x00),
	RGBA(0xFF, 0x00, 0x00, 0x00),
	RGBA(0xFF, 0x00, 0xFF, 0x00),
	RGBA(0x00, 0xFF, 0x00, 0x00),
	RGBA(0x00, 0xFF, 0xFF, 0x00),
	RGBA(0xFF, 0xFF, 0x00, 0x00),
	RGBA(0xFF, 0xFF, 0xFF, 0x00),
	RGBA(0x00, 0x00, 0x00, 0xFF),
	RGBA(0x00, 0x00, 0xFF, 0xFF),
	RGBA(0xFF, 0x00, 0x00, 0xFF),
	RGBA(0xFF, 0x00, 0xFF, 0xFF),
	RGBA(0x00, 0xFF, 0x00, 0xFF),
	RGBA(0x00, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0x00, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF)
};

static const struct bitmap_palette_t mask_palette[16] = {
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF)
};

static const struct bitmap_palette_t mono_palette[2] = {
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF)
};

static unsigned char shape_data[212544];
static unsigned char vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];
static unsigned char mask_vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];



static void usage(void);
static int draw_shape(int row, int col, const unsigned char *data);
static int write_shape_bmp(const char *name, int bit_count, const struct bitmap_palette_t *palette, const unsigned char *vram);
static int confirm(const char *msg);



static void usage(void)
{
	printf("usage: u4x68kshapes [options] OUTPUT.BMP [MASK.BMP]\n");
	printf("option: --disk DISK-IMG     disk image name of Ultima Disk\n");
	printf("                            read 'u4-ultima.xdf' if not specified\n");
	printf("        --help              print usage\n");
	printf("\n");
	printf("supported disk image format is *.xdf only\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *bmp_file = NULL;
	const char *mask_file = NULL;

	const char *disk_name = "u4-ultima.xdf";
	const char *file_name = "shape.pat";
	int row;
	int col;
	struct dosfd_image_t image;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (ultima disk)\n");
					return 1;
				}
				disk_name = *argv;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				bmp_file = *argv;
			} else {
				mask_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (bmp_file == NULL) {
		printf("no output bitmap filename\n");
		return 1;
	}

	if (read_dosfd_image(&image, disk_name, 0) < 0) {
		printf("read disk image error %s\n", disk_name);
		return 1;
	}
	if (read_dosfd_image_file(&image, file_name, shape_data, sizeof(shape_data), 1) < 0) {
		printf("read disk image error %s/%s\n", disk_name, file_name);
		return 1;
	}
	free_dosfd_image(&image, 0);

	memset(vram, 0, sizeof(vram));

	for (row = 0; row < NUM_ROWS; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			if (shape_info[row][col].shape < 0) {
				continue;
			}

			draw_shape(row, col, shape_data + shape_info[row][col].offset);
		}
	}

	if (write_shape_bmp(bmp_file, 4, palette, vram) < 0) {
		printf("write bitmap file error '%s'\n", bmp_file);
		return 1;
	}

	if (mask_file != NULL) {
		unsigned char *ptr = mask_vram;
		memset(mask_vram, 0, sizeof(mask_vram));
		if (convert_bit_count(SHAPE_WIDTH * NUM_COLS, SHAPE_HEIGHT * NUM_ROWS,
							  4, mask_palette, vram,
							  1, mono_palette, &ptr) < 0) {
			printf("convert error\n");
			return 1;
		}

		if (write_shape_bmp(mask_file, 1, mono_palette, mask_vram) < 0) {
			printf("write bitmap file error '%s'\n", mask_file);
			return 1;
		}
	}

	return 0;
}

static int draw_shape(int row, int col, const unsigned char *data)
{
	int y;
	int x;

	for (y = 0; y < SHAPE_HEIGHT; y++) {
		unsigned char *des = vram + (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;

		for (x = 0; x < SHAPE_WIDTH; x += 2) {
			unsigned char val = *data++;

			*des++ = val >> 4;
			*des++ = val & 0x0F;
		}
	}

	return 0;
}

static int write_shape_bmp(const char *name, int bit_count, const struct bitmap_palette_t *palette, const unsigned char *vram)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct stat st;

	if (stat(name, &st) == 0) {
		printf("already exists '%s'\n", name);
		if (confirm("overwrite ?") <= 0) {
			goto ERROR_EXIT;
		}
	}

	if ((fp = fopen(name, "wb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	create_bitmap_header(&header, SHAPE_WIDTH * NUM_COLS, SHAPE_HEIGHT * NUM_ROWS, bit_count);
	if (write_bitmap(fp, &header, palette, vram) < 0) {
		printf("write error '%s'\n", name);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
